"""Module for basic functionalities to operate with local file systems."""

import errno
import os

from py4ai.core.types import PathLike


def mkdir(path: PathLike) -> None:
    """
    Create a dir, using a formulation consistent between 2.x and 3.x python versions.

    :param path: path to create
    :raises OSError: whenever OSError is raised by makedirs and it's not because the directory exists
    """
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def create_dir_if_not_exists(directory: PathLike) -> PathLike:
    """
    Create a directory if it does not exist.

    :param directory: path
    :return: directory, str
    """
    if not os.path.exists(directory):
        os.makedirs(directory)
    return directory


def get_lexicographic_dirname(dirpath: PathLike, first: bool = False) -> PathLike:
    """
    Return the first (or last) subdirectory name ordered lexicographically.

    :param dirpath: name of the base path whose subdirectories ought to be listed
    :param first: whether the first or the last element should be returned
    :return: first (or last) subdirectory name ordered lexicographically.
    """
    return sorted(
        [
            os.path.join(dirpath, o).split("/")[-1]
            for o in os.listdir(dirpath)
            if os.path.isdir(os.path.join(dirpath, o))
        ],
        key=str.lower,
    )[0 if first else -1]



"""Implementation of classes that parse configuration files."""
import os
import re
from functools import reduce
from glob import glob
from typing import Optional, Sequence, Union

from cfg_load import Configuration, load
from yaml import (
    FullLoader,
    Loader,
    Node,
    SequenceNode,
    UnsafeLoader,
    add_constructor,
    add_implicit_resolver,
)

from py4ai.core.types import PathLike

env_var_matcher = re.compile(r"\$\{([^}^{]+)\}")


def path_constructor(
    loader: Union[Loader, FullLoader, UnsafeLoader], node: Node
) -> PathLike:
    """
    Extract the matched value, expand env variable, and replace the match.

    :param loader: not used
    :param node: YAML node
    :return: path
    :raises SyntaxError: if the node value does not match the regex expression for a path-like string
    :raises KeyError: raises an exception if the environment variable is missing
    """
    value = node.value
    match = env_var_matcher.match(value)

    if match is None:
        raise SyntaxError("Can't match pattern")

    env_var = match.group()[2:-1]
    try:
        return os.environ[env_var] + value[match.end() :]
    except KeyError:
        raise KeyError(
            f"Missing definition of environment variable {env_var} "
            f"needed when parsing configuration file"
        )


def joinPath(loader: Union[Loader, FullLoader, UnsafeLoader], node: Node) -> PathLike:
    """
    Join pieces of a file system path. Can be used as a custom tag handler.

    :param loader: YAML file loader
    :param node: YAML node
    :return: path
    :raises TypeError: if node is not a SequenceNode
    """
    if not isinstance(node, SequenceNode):
        raise TypeError("node input must be a SequenceNode")
    seq = loader.construct_sequence(node)
    return os.path.join(*seq)


# register tag handlers
add_implicit_resolver("!path", env_var_matcher)
add_constructor("!path", path_constructor)
add_constructor("!joinPath", joinPath)


def load_from_file(filename: PathLike) -> Configuration:
    """
    Load configuration reading given filename.

    :param filename: file to read
    :return: loaded configuration
    """
    try:
        return load(filename, safe_load=False, Loader=Loader)
    except TypeError:
        return load(filename)


def get_confs_in_path(path: PathLike, filename: str = "*.yml") -> Sequence[str]:
    """
    Retrieve all configuration files from system path, including the one in environment variable.

    :param path: path to search
    :param filename: filename can be either absolute (like /usr/src/Python-1.5/Makefile) or
        relative (like ../../Tools/*/*.gif), and can contain shell-style wildcards
    :return: list of retrieved configuration files paths
    """
    return [file for file in glob(os.path.join(path, filename))]


def merge_confs(
    filenames: Sequence[PathLike], default: Optional[str] = "defaults.yml"
) -> Configuration:
    """
    Merge configurations in given files.

    :param filenames: files to merge
    :param default: default configurations
    :return: merged configuration
    """
    lst = [default, *filenames] if default is not None else filenames
    print(f"Using Default Configuration file: {lst[0]}")
    return reduce(
        lambda config, fil: config.update(load_from_file(fil)),
        lst[1:],
        load_from_file(lst[0]),
    )



"""Typing utilities."""
import os
from typing import Any, Hashable, TypeVar, Union

from pydantic import BaseModel
from typing_extensions import Protocol

PathLike = Union[str, "os.PathLike[str]"]
T = TypeVar("T")
T_co = TypeVar("T_co", covariant=True)

# To be used as hashable key
K = TypeVar("K", bound=Hashable)
KE = TypeVar("KE")  # entity key
KD = TypeVar("KD")  # data key
E = TypeVar("E", bound=BaseModel)  # entity model
D = TypeVar("D")  # data model
Q = TypeVar("Q")  # search query
A = TypeVar("A")  # aggregate criteria
S = TypeVar("S", bound=BaseModel)  # generic serialized model
U = TypeVar("U")  # for unit of work usage


class SupportsLessThan(Protocol):
    """Protocol for a class that must support the less than operator."""

    def __lt__(self, other: Any) -> bool:
        """
        Less than operator.

        :param other: other operand
        :return bool: whether self is less than other
        """
        ...



"""Basic utilities extending pandas functionalities."""

from typing import Any, List, Union

from pandas import DataFrame, Index
from pandas.core.arrays.sparse.dtype import SparseDtype


def is_sparse(df: DataFrame) -> bool:
    """
    Return `True` if a dataframe is sparse.

    :param df: pandas dataframe
    :return: bool
    """
    return all([isinstance(v, SparseDtype) for k, v in df.dtypes.items()])


def loc(df: DataFrame, idx: Union[Index, List[Any]]):
    """
    Return a filtered dataframe based on an index list: it is designed to work properly with sparse dataframe as well.

    :param df: DataFrame
    :param idx: index list
    :return: DataFrame which is filtered
    """
    if is_sparse(df):
        csr = df.sparse.to_coo().tocsr()
        pos = [pos for pos, elem in enumerate(df.index) if elem in idx]
        return DataFrame.sparse.from_spmatrix(
            csr[pos, :],
            index=idx[pos]
            if isinstance(idx, Index)
            else [elem for elem in df.index if elem in idx],
            columns=df.columns,
        )
    else:
        return df.loc[idx]



"""Module for basic dictionary functionalities."""

import sys

if sys.version_info[0] < 3:
    from itertools import izip as zip
else:
    from functools import reduce

from collections.abc import Mapping
from copy import deepcopy as copy
from itertools import groupby, islice, tee
from operator import add
from typing import Any, Callable, Dict, Iterable, Iterator, List, Optional, Tuple

from py4ai.core.types import SupportsLessThan, T


def groupIterable(iterable: Iterable[T], batch_size: int = 10000) -> Iterator[List[T]]:
    """
    Split a given iterable into batches of given `batch_size`.

    :param iterable: iterable
    :param batch_size: int
    :return: Iterator
    """
    iterable = iter(iterable)
    return iter(lambda: list(islice(iterable, batch_size)), [])


def pairwise(iterable: Iterable[T]) -> zip:
    """
    Return a pairing of elements of an iterable.

    Example: s -> (s0,s1), (s1,s2), (s2, s3), ...

    :param iterable: iterable whose elements ought to be paired
    :return: zipped iterable
    """
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


def union(*dicts: dict) -> dict:
    """
    Return a dictionary that results from the recursive merge of the input dictionaries.

    :param dicts: list of dicts
    :return: merged dict
    """

    def __dict_merge(dct: dict, merge_dct: dict):
        """
        Recursive dict merge.

        Inspired by :meth:``dict.update()``, instead of updating only top-level keys, dict_merge recurses down into
        dicts nested to an arbitrary depth, updating keys. The ``merge_dct`` is merged into ``dct``.

        :param dct: dict onto which the merge is executed
        :param merge_dct: dct merged into dct
        :return: None
        """
        merged = copy(dct)
        for k, v in merge_dct.items():
            if (
                k in dct
                and isinstance(dct[k], dict)
                and isinstance(merge_dct[k], Mapping)
            ):
                merged[k] = __dict_merge(dct[k], merge_dct[k])
            else:
                merged[k] = merge_dct[k]
        return merged

    return reduce(__dict_merge, dicts)


def flattenKeys(input_dict: Dict[str, Any], sep: str = ".") -> Dict[str, Any]:
    """
    Flatten dictionary keys.

    Return a dictionary whose nested keys have been flattened into the root level recursively flattening a multilevel
    nested dict into a single level dict with flattened keys and corresponding values.
    The keys are joined by the given separator.

    :param input_dict: a multilevel dict ex `{"a": {"b": {"c": 2}}, "d": 2, "e": 3}`
    :param sep: str delimiter to join keys
    :return: dict with flattened keys
    """

    def _flatten_(key: str, value: Any) -> List[Tuple[str, Any]]:
        if isinstance(value, dict) and (len(value) > 0):
            return reduce(
                add,
                [
                    _flatten_(sep.join([key, name]), item)
                    for name, item in value.items()
                ],
            )
        else:
            return [(key, value)]

    return union(*[dict(_flatten_(key, value)) for key, value in input_dict.items()])


def unflattenKeys(input_dict: Dict[str, Any], sep: str = ".") -> Dict[str, Any]:
    """
    Transform a dict into a nested dict by splitting keys on the given separator.

    A dict with key as `{'a.b':2}` will be transformed into a dict of dicts as `{'a':{'b':2}}`

    :param input_dict: dict
    :param sep: str delimiter to split keys
    :return: dict
    """

    def _translate(key: str, value: Any) -> Dict[str, Any]:
        levels = list(reversed(key.split(sep)))
        return reduce(lambda agg, level: {level: agg}, levels[1:], {levels[0]: value})

    return union(*[_translate(key, value) for key, value in input_dict.items()])


def _check(value: Optional[T]) -> bool:
    return False if value is None else True


def filterNones(_dict: Dict[T, Any]) -> Dict[T, Any]:
    """
    Return a dictionary where the key,value pairs are filtered where the value is None.

    :param _dict: dict with Nones
    :return: dict without Nones
    """
    agg = {}
    for k, v in _dict.items():
        if isinstance(v, dict):
            agg[k] = filterNones(v)
        elif _check(v):
            agg[k] = v
    return agg


def groupBy(
    lst: Iterable[T], key: Callable[[T], SupportsLessThan]
) -> Iterator[Tuple[SupportsLessThan, List[T]]]:
    """
    Perform groupBy operation on a list according to the given key.

    The function uses itertools groupby but on a sorted list.

    :param lst: List
    :param key: function to be used as a key to perform groupby operation
    :yield: Iterator
    """
    for k, it in groupby(sorted(lst, key=key), key=key):
        yield k, list(it)



import os
import setuptools
import versioneer

with open(os.path.join("requirements", "requirements.in"), "r") as fid:
    requirements = [
        line.replace("\n", "")
        for line in fid.readlines()
        if not line.strip(" ").startswith("#")
    ]

setuptools.setup(
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    install_requires=requirements,
)
